import ast

import numpy as np
from collections import OrderedDict

from sklearn import model_selection
from sklearn.decomposition import PCA, LatentDirichletAllocation
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier, AdaBoostClassifier, ExtraTreesClassifier, IsolationForest
import sklearn.gaussian_process as gaussian_process
from sklearn.linear_model import Lasso, Ridge, ElasticNet, LinearRegression, LogisticRegression, SGDRegressor
from sklearn.kernel_ridge import KernelRidge
from sklearn.naive_bayes import MultinomialNB
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.svm import SVC, SVR
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier

from xgboost import XGBClassifier, XGBRegressor
from lightgbm import LGBMClassifier, LGBMRegressor
import catboost


def oof_generator(self, model, X_train, y_train, X_valid, n_folds=10):
    """
    Documentation:
        Description:
            Generates out-of-fold (OOF) predictions using provided model, training dataset
            and corresponding labels, and a validation dataset.
        Parameters:
            model : sklearn model or pipeline
                Model to be fit.
            X_train : array
                Training data observations.
            y_train : array
                Training target data.
            X_valid : array
                Validation data observations.
            n_folds : int, default=10
                Number of folds for performing cross-validation. Function will generate this
                many sets of out-of-fold predictions.
        Returns:
            oof_train : array
                Array containing observation data to be passed to the meta_learner. oof_train
                is updated throughout the cross-validation process with observations that are
                identified as test observations in each fold.
            oof_valid : array
                Array containing average of all out-of-fold predictions. This is the object to
                pass to the meta_learner.
    """
    # row counts
    n_train = X_train.shape[0]
    n_valid = X_valid.shape[0]

    # KFold train/test index generator
    kf = KFold(n_splits=n_folds)

    # create shell arrays for holding results
    oof_train = np.zeros((n_train,))
    oof_valid = np.zeros((n_valid,))
    oof_valid_scores = np.empty((n_folds, n_valid))

    # iterate through all KFolds to train model, capture scores
    for i, (train_ix, test_ix) in enumerate(kf.split(X_train)):
        
        # set train/test observations based on KFold indices
        X_train_fold = X_train[train_ix]
        y_train_fold = y_train[train_ix]
        x_test_fold = X_train[test_ix]

        # train model based on training variables and labels
        model.train(X_train_fold, y_train_fold)

        # update segment of oof_train where indices match the indices of the observations
        # used as test observations. these are the "out of fold" observations that we are not
        # considered in the training phase of the model
        oof_train[test_ix] = model.predict(x_test_fold)

        # generate predictions using entire validation dataset, otherwise unused up to this point
        # and capture predictions for each folds
        oof_valid_scores[i, :] = model.predict(X_valid)

    # determine average score of validation predictions
    oof_valid[:] = oof_valid_scores.mean(axis=0)
    return oof_train.reshape(-1, 1), oof_valid.reshape(-1, 1)


def model_stacker(self, models, bayes_optim_summary, X_train, y_train, X_valid, n_folds, n_jobs):
    """
    Documentation:
        Description:
            Stacking helper function.
        Parameters:
            models : dictionary
                Dictionary of 'x: y' pairs  to be fit.
            bayes_optim_summary : Pandas DataFrame
                Pandas DataFrame containing results from bayesian optimization process.
            X_train : array
                Training data observations.
            y_train : array
                Training target data.
            X_valid : array
                Validation data observations.
            n_folds : int
                Number of folds for performing cross-validation. Function will generate this
                many sets of out-of-fold predictions.
            n_jobs : int
                Number of workers to use when training the model. This parameter will be
                ignored if the model does not have this parameter.
        Returns:
            oof_train : array
                Out-of-fold training data observations.
            oof_valid : array
                Out-of-fold validation data observations.
            columns : list
                List containing estimator names.
    """
    columns = []

    # iterate through estimators
    for estimator in models.keys():

        # iterate through parameter set for estimator
        for model_iter in models[estimator]:
            print(estimator + " " + str(model_iter))
            columns.append(estimator + "_" + str(model_iter))

            model = self.BayesOptimModelBuilder(
                bayes_optim_summary=bayes_optim_summary,
                estimator_class=estimator,
                model_iter=model_iter,
                n_jobs=n_jobs,
            )

            oof_train_model, oof_valid_model = self.oof_generator(
                model=model,
                X_train=X_train,
                y_train=y_train,
                X_valid=X_valid,
                n_folds=n_folds,
            )
            try:
                oof_train = np.hstack((oof_train, oof_train_model))
                oof_valid = np.hstack((oof_valid, oof_valid_model))
            except NameError:
                oof_train = oof_train_model
                oof_valid = oof_valid_model
    return oof_train, oof_valid, columns



import numpy as np
import pandas as pd

from sklearn.base import TransformerMixin, BaseEstimator

from collections import Counter

import eif


class OutlierIQR(TransformerMixin, BaseEstimator):
    """
    Documentation:

        ---
        Description:
            Identifies outliers using inter quartile range (IQR) method using object
            oriented approach.

        ---
        Parameters:
            outlier_count : int
                Minimum number of values across all features for an single observation that
                need to be outliers in order for an observation to be flagged.
            iqr_step : float
                Multiplier that controls level of sensitivity of outlier detection method.
                Higher values for iqr_step will cause OutlierIQR to only detect increasingly
                extreme values.
            features : list
                List of features to be evaluated for outliers.
            drop_outliers : bool, default=False
                If True, drops outliers from the input data.

        ---
        Returns:
            x : array
                Dataset with outlier observations removed.
    """

    def __init__(self, outlier_count, iqr_step, features, drop_outliers=False):
        self.outlier_count = outlier_count
        self.iqr_step = iqr_step
        self.features = features
        self.drop_outliers = drop_outliers

    def fit(self, X, y=None):
        return self

    def transform(self, X):

        # empty list for holding index values for observations flagged as outlier
        outlier_indices = []

        # iterate over features
        for col in self.features:

            # calculate 25th percentile, 75th percentile and IQR
            q1 = np.percentile(X[col], 25)
            q3 = np.percentile(X[col], 75)
            iqr = q3 - q1

            # apply outlier step multiplier
            outlier_step = self.iqr_step * iqr

            # identify values that are higher or lower than the IQR
            outlier_list_col = X[
                (X[col] < q1 - outlier_step) | (X[col] > q3 + outlier_step)
            ].index

            # append the outlier observation index values
            outlier_indices.extend(outlier_list_col)

        ## identify observations with >= outlier_count outlier values among the features
        # count number of times each observation index appears in outlier_indices
        outlier_indices = Counter(outlier_indices)

        # reduce outlier_indices to the observations with >= outlier_count outlier values
        self.outliers = list(
            k for k, v in outlier_indices.items() if v >= self.outlier_count
        )

        # optionally drop outlier observations from input dataset
        if self.drop_outliers:
            X = X.drop(self.outliers, axis=0).reset_index(drop=True)

        return X


class ExtendedIsoForest(TransformerMixin, BaseEstimator):
    """
    Documentation:

        ---
        Description:
            Identifies outliers using extended isolation forest method.

        ---
        Parameters:
            columns : list
                Features to be evaluate.
            n_trees : int
                Number of trees used.
            sample_size : int
                Sub_sample size for creating each trees. Values must be smaller that input dataset
                row count.
            ExtensionLevel : int
                Degrees of freedom for choosing hyperplanes that divide data. Value must be smaller
                than input dataset column count.
            anomalies_ratio : float
                Percent of input dataset observations to identify as outliers.
            drop_outliers : bool, default=False
                Dictates whether identified outliers are removed from input dataset.

        ---
        Returns:
            x : array
                Dataset with outlier observations removed.
    """

    def __init__(self, columns, n_trees, sample_size, extension_level, anomalies_ratio, drop_outliers=False):
        self.columns = columns
        self.n_trees = n_trees
        self.sample_size = sample_size
        self.extension_level = extension_level
        self.anomalies_ratio = anomalies_ratio
        self.drop_outliers = drop_outliers

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        # instantiate extended isolation forest object
        ext_iso = eif.iForest(
            X=X[self.columns].values,
            ntrees=self.n_trees,
            sample_size=self.sample_size,
            ExtensionLevel=self.extension_level,
        )

        # calculate anomaly scores
        anomaly_scores = ext_iso.compute_paths(X_in=X[self.columns].values)

        # store anomaly score for each observation in Pandas DataFrame, sort descending
        anomaly_scores_sorted = pd.DataFrame(
            anomaly_scores, index=X.index, columns=["anomaly score"]
        ).sort_values(["anomaly score"], ascending=False)

        # identify outliers by limiting anomaly_scores_sorted to the (anamalies_ratio * total observation)
        # observations with the highest anomaly scores
        self.outliers = np.array(
            anomaly_scores_sorted[
                : int(np.ceil(self.anomalies_ratio * X.shape[0]))
            ].index
        )

        # optionally drop outlier observations from input dataset
        if self.drop_outliers:
            X = X.drop(self.outliers, axis=0).reset_index(drop=True)

        return X

def outlier_summary(self, iqr_outliers, if_outliers, eif_outliers):
    """
    Documentation:

        ---
        Description:
            Creates Pandas DataFrame summarizing which observations were flagged
            as outliers and by which outlier detection method each observation was
            identified.

        ---
        Parameters:
            iqr_outliers : array
                Array contains indexes of observations identified as outliers using
                IQR method.
            if_outliers : array
                Array contains indexes of observations identified as outliers using
                isolition forest method.
            eif_outliers : array
                Array contains indexes of observations identified as outliers using
                extended isolition forest method.

        ---
        Returns:
            outlier_summary : Pandas DataFrame
                DataFrame summarizing outlier detection results
    """
    # merge and de_duplicate outlier index values
    outlier_ixs = np.unique(np.concatenate([iqr_outliers, if_outliers, eif_outliers]))

    # create shell dataframe
    outlier_summary = pd.DataFrame(columns=["iqr", "if", "eif"], index=outlier_ixs)

    # fill nulls based on index value match
    outlier_summary["iqr"] = outlier_summary["iqr"].loc[iqr_outliers].fillna(value="x")
    outlier_summary["if"] = outlier_summary["if"].loc[if_outliers].fillna(value="x")
    outlier_summary["eif"] = outlier_summary["eif"].loc[eif_outliers].fillna(value="x")

    # add summary columns and sort
    outlier_summary["count"] = outlier_summary.count(axis=1)
    outlier_summary = outlier_summary.sort_values(["count"], ascending=False)

    outlier_summary = outlier_summary.fillna("")
    return outlier_summary

def outlier_IQR(self, data, iqr_step):
    """
    Documentation:

        ---
        Description:
            Identifies outliers using inter quartile range (IQR) method using functional
            approach.

        ---
        Parameters:
            data : Pandas Series
                Input data array.
            iqr_step : float
                Multiplier that controls level of sensitivity of outlier detection method.
                Higher values for iqr_step will cause OutlierIQR to only detect increasingly
                extreme values.

        ---
        Returns:
            outlier_index : array
                Index of outliers in original data array.
    """
    # calculate 25th percentile, 75th percentile and IQR
    q1 = np.percentile(data, 25)
    q3 = np.percentile(data, 75)
    iqr = q3 - q1

    # apply outlier step multiplier
    outlier_step = iqr_step * iqr

    # identify values that are higher or lower than the IQR
    outlier_index = data[(data < q1 - outlier_step) | (data > q3 + outlier_step)].index

    return outlier_index


from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
import pandas as pd
import numpy as np


class PowerGridSearcher:
    """
    Documentation:

        ---
        Definition:
            Class capable of performing GridSearchCV and RandomizedSearchCV on
            multiple models, each with its own parameter grid, in a single execution.
            Also returns a score summary sheet.

        ---
        Parameters:
            models : dict
                Dictionary of instantiated sklearn-style models.
            params : dict of dicts
                Dictionary containing nested dictionaries of parameter grids for each model.
    """

    def __init__(self, models, params):
        if not set(models.keys()).issubset(set(params.keys())):
            missing_params = list(set(models.keys()) - set(params.keys()))
            raise ValueError(
                "some estimators are missing Parameters: {0}".format(missing_params)
            )
        self.models = models
        self.params = params
        self.keys = models.keys()
        self.grid_searches = {}

    # full GridSearchCV
    def fit_multi_gs_cv(self, x, y, cv=5, n_jobs=1, verbose=0, scoring=None, refit=True):
        """
        Documentation:

            ---
            Definition:
                Method for performing GridSearchCV on multiple models, each with its own parameter
                grid, in a single execution.

            ---
            Parameters:
                x : array
                    Training data observation.
                y : array
                    Training target data.
                cv : int, default=5
                    The number of folds to use in cross-validation procedure.
                n_jobs : int, default=1
                    Number of workers to deploy upon execution, if applicable.
                verbose : int, default=0
                    Controls amount of information printed to console during fit.
                scoring : str (sklearn evaluation metric), default=None
                    Scoring metric to use in evaluation.
                refit : bool, default=True
                    Dictates whether method is refit with the best model upon grid seach completion.

            ---
            Returns:
                gs : GridSearchCV object
                    Full GridSearchCV object.
        """
        # iterate through all estimators
        for key in self.keys:
            print("running GridSearchCV for {0}".format(key))

            # separate current model and associated param grid diciontary
            model = self.models[key]
            params = self.params[key]

            # instantiate GridSearchCV object and fit
            gs = GridSearchCV(
                model,
                params,
                cv=cv,
                n_jobs=n_jobs,
                verbose=verbose,
                scoring=scoring,
                refit=refit,
                return_train_score=True,
            )
            gs.fit(x, y)

            # store GridSearchCV object
            self.grid_searches[key] = gs
        return gs

    # RandomizedSearchCV
    def fit_multi_rgs_cv(self, x, y, cv=5, n_jobs=1, verbose=0, scoring=None, refit=True, n_iter=15):
        """
        Documentation:

            ---
            Definition:
                method for performing randomized_GridSearchCV on multiple models, each with its own parameter
                grid, in a single execution.

            ---
            Parameters:
                x : array
                    Training data observation.
                y : array
                    Training target data.
                cv : int, default=5
                    The number of folds to use in cross-validation procedure.
                n_jobs : int, default=1
                    Number of workers to deploy upon execution, if applicable.
                verbose : int, default=0
                    Controls amount of information printed to console during fit.
                scoring : str (sklearn evaluation metric), default=None
                    Scoring metric to use in evaluation.
                refit : bool, default=True
                    Dictates whether method is refit with the best model upon grid seach completion.
                n_iter : int, default=15
                    Number of random permutations to evaluate.

            ---
            Returns:
                rgs : randomized_GridSearchCV object
                    Full randomized_GridSearchCV object.
        """
        # iterate through all estimators
        for key in self.keys:
            print("running RandomizedSearchCV for {0}".format(key))

            # separate current model and associated param grid diciontary
            model = self.models[key]
            params = self.params[key]

            # instantiate GridSearchCV object and fit
            rgs = RandomizedSearchCV(
                model,
                params,
                cv=cv,
                n_jobs=n_jobs,
                verbose=verbose,
                scoring=scoring,
                refit=refit,
                return_train_score=True,
                n_iter=n_iter,
            )
            rgs.fit(x, y)

            # store GridSearchCV object
            self.grid_searches[key] = rgs
        return rgs

    def score_summary(self, sort_by="mean_score"):
        """
        Documentation:

            ---
            Definition:
                Method for performing RandomizedGridSearchCV on multiple models, each with
                its own parameter grid, in a single execution.

            ---
            Parameters:
                sort_by : str (columns of score summary), default="mean_score"
                    Column on which to sort score summary.

            ---
            Returns:
                df : Pandas DataFrame
                    DataFrame containing results summary of GridSearchCV or
                    RandomizedGridSearchCV.
        """

        # helper function for collecting single row of summary data
        def row(key, scores, params):
            d = {
                "estimator": key,
                "min_score": min(scores),
                "max_score": max(scores),
                "mean_score": np.mean(scores),
                "std_score": np.std(scores),
            }
            return pd.Series({**params, **d})

        rows = []

        # iterate through grid search object
        for k in self.grid_searches:
            params = self.grid_searches[k].cv_results_["params"]
            scores = []
            for i in range(self.grid_searches[k].cv):
                key = "split{}_test_score".format(i)
                r = self.grid_searches[k].cv_results_[key]
                scores.append(r.reshape(len(params), 1))

            all_scores = np.hstack(scores)
            for p, s in zip(params, all_scores):
                rows.append((row(k, s, p)))

        df = pd.concat(rows, axis=1).t.sort_values([sort_by], ascending=False)

        columns = ["estimator", "min_score", "mean_score", "max_score", "std_score"]
        columns = columns + [c for c in df.columns if c not in columns]

        return df[columns]


def PowerGridModelBuilder(self, results, model_ix):
    """
    Documentation:
        
        ---
        Description:
            Model building utility to be used with power_grid_search summary results.
            Returns estimator and dictionary of 'parameter : value' pairs to be passed
            as kwarg to model. Used to efficiently reinstantiated specific models.
        
        ---
        Parameters:
            results : Pandas DataFrame
                DataFrame containing score summary generated by PowerGridSearcher.
            model_ix : int
                Row index of specific model summarized in results DataFrame.
        
        ---
        Returns:
            estimator : sklearn model
                sklearn model object.
            params.to_dict() : dictionary
                Dictionary containing model parameters value.
    """
    estimator = results.loc[model_ix][0]
    params = results.loc[model_ix][5:].dropna(axis=0)

    # convert floats that are effectively ints to ints
    for ix in params.index:
        if not isinstance(params[ix], str):
            if int(params[ix]) == params[ix] and isinstance(params[ix], float):
                params[ix] = params[ix].astype(np.int64)

    return estimator, params.to_dict()



import numpy as np

import matplotlib.pyplot as plt

from scipy import stats
from scipy import special

from prettierplot.plotter import PrettierPlot
from prettierplot import style


def eda_missing_summary(self, data=None, color=style.style_grey, display_df=False, chart_scale=15):
    """
    Documentation:

        ---
        Description:
            Creates vertical bar chart visualizing the percent of values missing for each feature.
            Optionally displays the underlying Pandas DataFrame.

        ---
        Parameters:
            data : Pandas DataFrame, default=None
                Pandas DataFrame containing independent variables. If left as none,
                the feature dataset provided to Machine during instantiation is used.
            color : str or color code, default=style.style_grey
                Bar color.
            display_df : boolean, default=False
                Controls whether to display summary data in Pandas DataFrame in addition to chart.
            chart_scale : int or float, default=15
                Controls size and proportions of chart and chart elements. Higher value creates
                larger plots and increases visual elements proportionally.
    """
    # use dataset provided during instantiation if None
    if data is None:
        data = self.data

    # return missingness summary
    percent_missing = self.missing_summary(data)

    # if missingness summary is not empty, create the visualization
    if not percent_missing.empty:
        # optionally display DataFrame summary
        if display_df:
            display(percent_missing)

        # create prettierplot object
        p = PrettierPlot(chart_scale=chart_scale, plot_orientation="wide_standard")

        # add canvas to prettierplot object
        ax = p.make_canvas(
            title="Percent missing by feature",
            y_shift=0.8,
            title_scale=0.8,
        )

        # add vertical bar chart to canvas
        p.bar_v(
            x=percent_missing.index,
            counts=percent_missing["Percent missing"],
            label_rotate=45 if len(percent_missing.index) <=5 else 90,
            color=color,
            y_units="p",
            x_tick_wrap=False,
            ax=ax,
        )
    
    # if missingness summary is empty, just print "No Nulls"
    else:
        print("No nulls")

def eda_skew_summary(self, data=None, color=style.style_grey, display_df=False, chart_scale=15):
    """
    Documentation:

        ---
        Description:
            Creates vertical bar chart visualizing the skew for each feature. Optionally
            displaying the underlying Pandas DataFrame.

        ---
        Parameters:
            data : Pandas DataFrame, default=None
                Pandas DataFrame containing independent variables. If left as none,
                the feature dataset provided to Machine during instantiation is used.
            color : str, color code, default=style.style_grey
                Bar color.
            display_df : boolean, default=False
                Controls whether to display summary data in Pandas DataFrame along with chart.
            chart_scale : int or float, default=15
                Controls size and proportions of chart and chart elements. Higher value creates
                larger plots and increases visual elements proportionally.
    """
    # use dataset provided during instantiation if None
    if data is None:
        data = self.data

    # return skewness summary
    skew_summary = self.skew_summary(data)

    # optionally display DataFrame summary
    if display_df:
        display(skew_summary)

    # create prettierplot object
    p = PrettierPlot(chart_scale=chart_scale, plot_orientation="wide_standard")

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="Skew by feature",
        y_shift=0.8,
        title_scale=0.8,
    )

    # add vertical bar chart to canvas
    p.bar_v(
        x=skew_summary.index,
        counts=skew_summary["Skew"],
        label_rotate=45 if len(skew_summary.index) <=5 else 90,
        color=color,
        y_units="fff",
        x_tick_wrap=False,
        ax=ax,
    )

def eda_transform_target(self, data, name, chart_scale=15):
    """
    Documentation:

        ---
        Description:
            Creates a two_panel visualization. The left plot is the current distribution
            overlayed on a normal distribution. The right plot is a qqplot overlayed 
            across a straight line.

        ---
        Parameters:
            data : Pandas Series
                Target variable data object.
            name : str
                Name of target variable.
            chart_scale : int or float, default=15
                Controls size and proportions of chart and chart elements. Higher value
                creates larger plots and increases visual elements proportionally.
    """
    # create prettierplot object
    p = PrettierPlot(chart_scale=chart_scale)

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="dist/kde - {} (initial)".format(name),
        x_label="",
        y_label="",
        y_shift=0.8,
        position=221,
    )

    # add distribution / kernel density plot to canvas
    p.dist_plot(
        data, color=style.style_grey, fit=stats.norm, x_rotate=True, ax=ax
    )

    # turn off x and y ticks
    plt.xticks([])
    plt.yticks([])

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="probability plot - {} (initial)".format(name),
        x_label="",
        y_label="",
        y_shift=0.8,
        position=222,
    )

    # add QQ / probability plot to canvas
    p.prob_plot(data, plot=ax)

    # turn off x and y ticks
    plt.xticks([])
    plt.yticks([])


def eda_transform_log1(self, data, name, chart_scale=15):
    """
    Documentation:

        ---
        Description:
            Creates a two_panel visualization. The left plot is the log + 1 transformed
            distribution overlayed on a normal distribution. The right plot is a log + 1 
            adjusted qqplot overlayed across a straight line.

        ---
        Parameters:
            data : Pandas Series
                Target variable data object.
            name : str
                Name of target variable.
            chart_scale : int or float, default=15
                Controls size and proportions of chart and chart elements. Higher value creates
                larger plots and increases visual elements proportionally.
    """
    # create prettierplot object
    p = PrettierPlot(chart_scale=chart_scale)

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="dist/kde - {} (log+1)".format(name),
        x_label="",
        y_label="",
        y_shift=0.8,
        position=223,
    )

    # add distribution / kernel density plot to canvas
    p.dist_plot(
        np.log1p(data), color=style.style_grey, fit=stats.norm, x_rotate=True, ax=ax
    )

    # turn off x and y ticks
    plt.xticks([])
    plt.yticks([])

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="probability plot - {} (log+1)".format(name),
        x_label="",
        y_label="",
        y_shift=0.8,
        position=224,
    )

    # add QQ / probability plot to canvas
    p.prob_plot(np.log1p(data), plot=ax)

    # turn off x and y ticks
    plt.xticks([])
    plt.yticks([])


def eda_transform_box_cox(self, data, name, lmbda, chart_scale=15):
    """
    Documentation:

        ---
        Description:
            Creates a two_panel visualization. The left plot is the box-cox transformed
            distribution overlayed on a normal distribution. The right plot is a Box-Cox
            transformed qqplot overlayed across a straight line.

        ---
        Parameters:
            data : Pandas Series
                Target variable data object.
            name : str
                Name of target variable.
            lmbda : float
                Box-Cox transformation parameter.
            chart_scale : int or float, default=15
                Controls size and proportions of chart and chart elements. Higher value
                creates larger plots and increases visual elements proportionally.
    """
    # create prettierplot object
    p = PrettierPlot(chart_scale=chart_scale)

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="dist/kde - {} (box-cox, {})".format(name, lmbda),
        x_label="",
        y_label="",
        y_shift=0.8,
        position=223,
    )

    # add distribution / kernel density plot to canvas
    p.dist_plot(
        special.boxcox1p(data, lmbda),
        color=style.style_grey,
        fit=stats.norm,
        x_rotate=True,
        ax=ax,
    )

    # turn off x and y ticks
    plt.xticks([])
    plt.yticks([])

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="Probability plot - {} (box-cox, {})".format(name, lmbda),
        x_label="",
        y_label="",
        y_shift=0.8,
        position=224,
    )

    # add QQ / probability plot to canvas
    p.prob_plot(special.boxcox1p(data, lmbda), plot=ax)

    # turn off x and y ticks
    plt.xticks([])
    plt.yticks([])


import setuptools


# with open("README.md") as f:
#     long_description = f.read()

long_description = "mlmachine is a Python library that organizes and accelerates notebook-based machine learning experiments."

description = "Accelerate machine learning experimentation"
distname = "mlmachine"
license = "MIT"
# download_url = 'https://pypi.org/project/'
maintainer = "Tyler Peterson"
maintainer_email = "petersontylerd@gmail.com"
project_urls = {
    "bug tracker": "https://github.com/petersontylerd/mlmachine/issues",
    "source code": "https://github.com/petersontylerd/mlmachine",
}
url = "https://github.com/petersontylerd/mlmachine"
version = "0.1.1"


def setup_package():
    metadata = dict(
        name=distname,
        packages=[
            "mlmachine",
            "mlmachine.datasets",
            "mlmachine.datasets.attrition",
            "mlmachine.datasets.housing",
            "mlmachine.datasets.titanic",
            "mlmachine.explore",
            "mlmachine.features",
            "mlmachine.model",
            "mlmachine.model.evaluate",
            "mlmachine.model.explain",
            "mlmachine.model.tune",
        ],
        maintainer=maintainer,
        maintainer_email=maintainer_email,
        description=description,
        keywords=["machine learning", "data science"],
        license=license,
        url=url,
        # download_url = download_url,
        project_urls=project_urls,
        version=version,
        long_description=long_description,
        include_package_data=True,
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Scientific/Engineering :: Information Analysis",
            "Topic :: Scientific/Engineering :: Visualization",
            "Topic :: Software Development",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
        ],
        python_requires=">=3.6.1",
        install_requires=[i.strip() for i in open("requirements.txt").readlines()],
        # dependency_links=["https://github.com/petersontylerd/prettierplot"],
    )

    setuptools.setup(**metadata)


if __name__ == "__main__":
    setup_package()



import numpy as np
import pandas as pd

from pandas.api.types import (
    is_bool_dtype,
    is_categorical_dtype,
    is_datetime64_any_dtype,
    is_numeric_dtype,
    is_object_dtype,
    is_string_dtype,
    CategoricalDtype,
)

from scipy import sparse, stats, special

from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.impute import SimpleImputer
from sklearn.pipeline import (
    make_pipeline,
    Pipeline,
    FeatureUnion,
    _fit_transform_one,
    _transform_one,
)

from sklearn.preprocessing import (
    StandardScaler,
    RobustScaler,
    PolynomialFeatures,
    OrdinalEncoder,
    LabelEncoder,
    OneHotEncoder,
    KBinsDiscretizer,
    QuantileTransformer,
    PowerTransformer,
    MinMaxScaler,
    _encoders,
    _data,
    _discretization,
)
from sklearn.externals.joblib import Parallel, delayed

from category_encoders import WOEEncoder, TargetEncoder, CatBoostEncoder, BinaryEncoder, CountEncoder

import itertools
import collections
import copy


class GroupbyImputer(TransformerMixin, BaseEstimator):
    """
    Documentation:

        ---
        Description:
            Impute numeric columns, as specified by the strategy parameter. GroupbyImputer utilizes one or
            more additional context columns as a groupby value to add more subtlety to fill_value identification.
            Imputes training data features, and stores impute values to be used on validation and unseen data.

        ---
        Parameters:
            null_column : list
                Column with nulls to impute.
            groupby_column : list
                List of one or most columns to groupby to add context to null column.
            strategy : str, default='mean'
                Imputing stategy. Accepts values 'mean', 'median' and 'most_frequent'.
            train : bool, default=True
                Tells class whether we are imputing training data or unseen data.
            train_value : dict, default=None
                Only used when train=False. Value is a dictionary containing 'feature : value' pairs.
                Dictionary is retrieved from training data pipeline using named steps. The attribute is
                called train_value_.

        ---
        Returns:
            X : array
                Dataset where each column with missing values has been imputed with a value learned from a particular
                strategy while also consider select columns as a group by variable.
    """

    def __init__(self, null_column, groupby_column, strategy="mean", train=True, train_value=None):
        self.null_column = null_column
        self.groupby_column = groupby_column
        self.strategy = strategy
        self.train = train
        self.train_value = train_value

    def fit(self, X, y=None):

        # if imputation strategy is set to "mean"
        if self.strategy == "mean":

            # grouping by groupby_column, find mean of null_column
            self.train_value = (
                X[X[self.null_column].notnull()]
                .groupby(self.groupby_column)
                .mean()[self.null_column]
            )
            # calculate overall mean of null_column
            self.overall = X[X[self.null_column].notnull()][self.null_column].mean()

        # if imputation strategy is set to "median"
        elif self.strategy == "median":
            # grouping by groupby_column, find median of null_column
            self.train_value = (
                X[X[self.null_column].notnull()]
                .groupby(self.groupby_column)
                .median()[self.null_column]
            )
            # calculate overall median of null_column
            self.overall = X[X[self.null_column].notnull()][self.null_column].median()

        # if imputation strategy is set to "most_frequent"
        elif self.strategy == "most_frequent":
            # grouping by groupby_column, find mode of null_column
            self.train_value = (
                X[X[self.null_column].notnull()]
                .groupby(self.groupby_column)[self.null_column]
                .agg(lambda X: X.value_counts().index[0])
            )
            # calculate overall mode of null_column
            self.overall = X[X[self.null_column].notnull()][self.null_column].mode()[0]

        self.train_value = self.train_value.reset_index()

        return self

    def transform(self, X):
        # impute missing values based on train_value
        if isinstance(self.groupby_column, str):

            # impute nulls with corresponding value
            X[self.null_column] = np.where(
                X[self.null_column].isnull(),
                X[self.groupby_column].map(
                    self.train_value.set_index(self.groupby_column)[self.null_column]
                ),
                X[self.null_column],
            )

            # impute any remainig nulls with overall value
            X[self.null_column] = X[self.null_column].fillna(value=self.overall)

        return X[self.null_column]

class DataFrameSelector(BaseEstimator, TransformerMixin):
    """
    Documentation:

        ---
        Description:
            Select a susbset set of features from Pandas DataFrame as part of a
            pipeline. Capable of select and/or deselecting columns by name, Pandas
            dtype and mlm dtype.

            Note - if there is a logical conflict between include and exclude
            parameters, the class will first prioritize the column parameters
            over the dtype parameters in order to support subsetting. If the
            logic cannot be resolved by this rule alone, exclusion parameters
            will be prioritized over inclusion parameters.

        ---
        Parameters:
            include_columns : list
                List of features to select from Pandas DataFrame.
            include_pd_dtypes : list
                List of strings describing pandas dtypes to select.
            include_mlm_dtypes : list
                List of strings describing mlmachine dtypes to select.
            exclude_columns : list
                List of features to exclude from Pandas DataFrame.
            exclude_pd_dtypes : list
                List of strings describing pandas dtypes to exclude.
            exclude_mlm_dtypes : list
                List of strings describing mlmachine dtypes to exclude.
    """

    def __init__(self, include_columns=None, include_pd_dtypes=None, include_mlm_dtypes=None,
                    exclude_columns=None, exclude_pd_dtypes=None, exclude_mlm_dtypes=None):

        self.include_columns = include_columns
        self.include_pd_dtypes = include_pd_dtypes
        self.include_mlm_dtypes = include_mlm_dtypes
        self.exclude_columns = exclude_columns
        self.exclude_pd_dtypes = exclude_pd_dtypes
        self.exclude_mlm_dtypes = exclude_mlm_dtypes

    def fit(self, X, y=None):

        # preserve mlm_dtypes attribute on Pandas DataFrame, if it exists
        try:
            self.meta_mlm_dtypes = X.mlm_dtypes
            self.no_meta_mlm_dtypes = False
        except AttributeError:
            if self.include_mlm_dtypes is not None or self.exclude_mlm_dtypes is not None:
                raise AttributeError ("Attempting to filter using mlm dtypes, but mlm_dtypes object is not associated with the input Pandas DataFrame.")
            else:
                pass

        # capture all current columns in input Pandas DataFrame
        self.all_columns = X.columns.tolist()

        # empty list to capture selected columns
        self.selected_columns = []

        # empty list to capture excluded columns
        self.remove_columns = []

        ## selection
        # select columns by name
        if self.include_columns is not None:
            self.selected_columns.extend(self.include_columns)

        # select columns by pandas dtype
        if self.include_pd_dtypes is not None:
            for dtype in self.include_pd_dtypes:
                self.selected_columns.extend(
                    X.select_dtypes(include=dtype).columns.tolist()
                )

        # select columns by mlm dtype
        if self.include_mlm_dtypes is not None:
            for dtype in self.include_mlm_dtypes:
                self.selected_columns.extend(X.mlm_dtypes[dtype])

        # flatten list and remove duplicates
        self.selected_columns = collections.OrderedDict.fromkeys(self.selected_columns)
        self.selected_columns = list(self.selected_columns.keys())

        ## deselection
        # deselect columns by name
        if self.exclude_columns is not None:
            self.remove_columns.extend(self.exclude_columns)

        # deselect columns by pandas dtype
        if self.exclude_pd_dtypes is not None:
            for dtype in self.exclude_pd_dtypes:
                self.remove_columns.extend(
                    X.select_dtypes(include=dtype).columns.tolist()
                )

        # deselect columns by mlm dtype
        if self.exclude_mlm_dtypes is not None:
            for dtype in self.exclude_mlm_dtypes:
                self.remove_columns.extend(X.mlm_dtypes[dtype])

        # flatten list and remove duplicates
        self.remove_columns = list(set(self.remove_columns))

        ## reconcile selections and removals
        # if the only input is to exclude, remove remove_columns from all_columns
        if len(self.remove_columns) > 0 and len(self.selected_columns) == 0:
            # set final_column to all_columns minus the columns in remove_columns
            self.final_columns = list(
                set(self.all_columns).difference(self.remove_columns)
            )

        # if the only input is to include, keep only the select_columns
        elif len(self.selected_columns) > 0 and len(self.remove_columns) == 0:
            # set final_column to selected_columns
            self.final_columns = self.selected_columns

        # if columns are selected by name and columns are excluded by name
        elif (
            self.include_columns is not None
            and (self.include_pd_dtypes is None and self.include_mlm_dtypes is None)
            and self.exclude_columns is not None
            and (self.exclude_pd_dtypes is None and self.exclude_mlm_dtypes is None)
        ):
            # set final_column to selected_columns and remove all columns in remove_columns
            self.final_columns = self.selected_columns
            self.final_columns = [column for column in self.final_columns if column not in self.remove_columns]

        # if columns are selected by name and columns are excluded by pandas dtype or mlm dtype
        elif (
            self.include_columns is not None
            and (self.include_pd_dtypes is None and self.include_mlm_dtypes is None)
            and self.exclude_columns is None
            and (self.exclude_pd_dtypes is not None or self.exclude_mlm_dtypes is not None)
        ):
            # set final_column to selected_columns
            self.final_columns = self.selected_columns

        # if columns are excluded by name and columns are selected by pandas dtype or mlm dtype
        elif (
            self.include_columns is None
            and (self.include_pd_dtypes is not None or self.include_mlm_dtypes is not None)
            and self.exclude_columns is not None
            and (self.exclude_pd_dtypes is None and self.exclude_mlm_dtypes is None)
        ):
            # set final_column to selected_columns and remove all columns in remove_columns
            self.final_columns = self.selected_columns
            self.final_columns = [column for column in self.final_columns if column not in self.remove_columns]

        #
        # if columns are selected by pandas dtype or mlm dtype, and  columns are excluded by pandas dtype or mlm dtype
        elif (
            self.include_columns is None
            and (self.include_pd_dtypes is not None or self.include_mlm_dtypes is not None)
            and self.exclude_columns is None
            and (self.exclude_pd_dtypes is not None or self.exclude_mlm_dtypes is not None)
        ):
            # set final_column to selected_columns and remove all columns in remove_columns
            self.final_columns = self.selected_columns
            self.final_columns = [column for column in self.final_columns if column not in self.remove_columns]

        # if columns are selected by pandas dtype or mlm dtype
        elif (
            self.include_columns is not None
            and (self.include_pd_dtypes is not None or self.include_mlm_dtypes is not None)
            and self.exclude_columns is not None
            and (self.exclude_pd_dtypes is None and self.exclude_mlm_dtypes is None)
        ):
            # set final_column to selected_columns and remove all columns in remove_columns
            self.final_columns = self.selected_columns
            self.final_columns = [column for column in self.final_columns if column not in self.remove_columns]

        # if columns are selected by name, and selected by pandas dtype or mlm dtype, and columns are excluded
        # by pandas dtype or mlm dtype
        elif (
            self.include_columns is not None
            and (self.include_pd_dtypes is not None or self.include_mlm_dtypes is not None)
            and self.exclude_columns is None
            and (self.exclude_pd_dtypes is not None or self.exclude_mlm_dtypes is not None)
        ):
            # set final_column to selected_columns
            self.final_columns = self.selected_columns

        # if columns are excluded by name, and selected by pandas dtype or mlm dtype, and columns are excluded
        # by pandas dtype or mlm dtype
        elif (
            self.include_columns is None
            and (self.include_pd_dtypes is not None or self.include_mlm_dtypes is not None)
            and self.exclude_columns is not None
            and (self.exclude_pd_dtypes is not None or self.exclude_mlm_dtypes is not None)
        ):
            # set final_column to selected_columns and remove all columns in remove_columns
            self.final_columns = self.selected_columns
            self.final_columns = [column for column in self.final_columns if column not in self.remove_columns]

        # if columns are selected by name and excluded by name, and excluded by pandas dtype or mlm dtype
        elif (
            self.include_columns is not None
            and (self.include_pd_dtypes is None and self.include_mlm_dtypes is None)
            and self.exclude_columns is not None
            and (self.exclude_pd_dtypes is not None or self.exclude_mlm_dtypes is not None)
        ):
            # set final_column to selected_columns
            self.final_columns = self.selected_columns

        # if columns are selected by name and excluded by name, and selected by pandas dtype or mlm dtype, and
        # excluded by pandas dtype or mlm dtype
        elif (
            self.include_columns is not None
            and (self.include_pd_dtypes is not None or self.include_mlm_dtypes is not None)
            and self.exclude_columns is not None
            and (self.exclude_pd_dtypes is not None or self.exclude_mlm_dtypes is not None)
        ):
            # set final_column to selected_columns
            self.final_columns = self.selected_columns

        # if all parameters are left to default to None
        elif (
            self.include_columns is None
            and self.include_pd_dtypes is None
            and self.include_mlm_dtypes is None
            and self.exclude_columns is None
            and self.exclude_pd_dtypes is None
            and self.exclude_mlm_dtypes is None
        ):

            # set final_column to all_columns
            self.final_columns = self.all_columns

        return self

    def transform(self, X):
        return X[self.final_columns]

class PandasTransformer(TransformerMixin, BaseEstimator):
    """
    Documentation:

        ---
        Description:
            Wrapper that ensures sklearn transformers will play nicely with Pandas.
            Pipelines will output a Pandas DataFrame with column names and original
            index values rather than a two dimensional numpy array.

        ---
        Parameters
            transformer : sklearn transformer
                Transformer to apply to dataset
    """
    def __init__(self, transformer):
        self.transformer = transformer

    def fit(self, X, y=None):

        # capture original column names and dtypes
        self.original_columns = X.columns
        self.dtypes = X[self.original_columns].dtypes

        # create object containing fitted transformer
        self.est = self.transformer.fit(X)

        # if the class is a OneHotEncoder instance
        if isinstance(self.est, _encoders.OneHotEncoder):

            # capture feature names learned by OneHotEncoder
            names = self.est.get_feature_names()

            # transform names into readable column names
            prefixes = list(map(lambda x:str.replace(x, "x", ""), names))
            prefixes = sorted(list(set([x.split("_")[0] for x in prefixes])), key=int)
            prefixes = ["x" + x + "_" for x in prefixes]

            # add "_{category name}" suffix to all column names
            for a,b in zip(self.original_columns, prefixes):
                names = list(map(lambda x: str.replace(x, b, a + "_"), names))
            self.original_columns = names

        # if the class is a PolynomialFeatures instance
        elif isinstance(self.est, _data.PolynomialFeatures):

            ### replace feature code names with actual feature names
            # capture object's feature code names
            names = self.est.get_feature_names()

            # remove "x" prefix from features code names
            feature_code_names = list(map(lambda x:str.replace(x, "x", ""), names))

            # reduce feature_code_names down to [0,1,2,3...] by ignoring "^2" and sort as integers
            feature_code_names = sorted(list(set([x.replace("^2","").split(" ")[0] for x in feature_code_names])), key=int)

            # add "x" as a prefix and an "_" as a suffix to the
            feature_code_names = ["x" + x +"_" for x in feature_code_names]

            ### build actual feature names
            # replace " " with "_ " and add an additional "_" to the full string yield names such as "x0_ x1_"
            feature_actual_names = [x.replace(" ","_ ") + "_" for x in names]

            # change strings such as "x02_" to yield names such as "x0_2"
            feature_actual_names = [x.replace("^2_","_^2") for x in feature_actual_names]

            # change strings such as "x0_ x1_" to yield names such as "x0_*x1_"
            feature_actual_names = [x.replace(" ","*") for x in feature_actual_names]

            # replace code names with actual column names
            for a,b in zip(self.original_columns, feature_code_names):
                feature_actual_names = list(map(lambda x: str.replace(x, b, a), feature_actual_names))

            self.original_columns = feature_actual_names

        # if the class is a KBinsDiscretizer instance
        elif isinstance(self.est, KBinsDiscretizer):
            names = []

            # capture number of bins learned by KBinsDiscretizer
            bins = self.est.n_bins

            # add "_binned_{# of bins}" suffix to all column names
            for col in self.original_columns:
                names.append(col + "_binned_" + str(bins))

            self.original_columns = names

        # if the class is a OrdinalEncoder instance
        elif isinstance(self.est, OrdinalEncoder):
            names = []

            # add "_ordinal_encoded" suffix to all column names
            for col in self.original_columns:
                names.append(col + "_ordinal_encoded")

            self.original_columns = names

        # if the class is a category_encoders CountEncoder instance
        elif isinstance(self.est, CountEncoder):
            names = []

            # add "_count_encoded" suffix to all column names
            for col in self.original_columns:
                names.append(col + "_count_encoded")

            self.original_columns = names

        # if the class is a category_encoders BinaryEncoder instance
        elif isinstance(self.est, BinaryEncoder):
            names = []
            self.original_columns = self.est.get_feature_names()

            # add "_binary_encoded" suffix to all column names
            for col in self.original_columns:
                names.append(col + "_binary_encoded")

            self.original_columns = names

        # if the class is a category_encoders BinaryEncoder instance
        elif isinstance(self.est, QuantileTransformer):
            names = []

            # add "_quantile_{distribution type}" suffix to all column names
            for col in self.original_columns:
                names.append(col + "_quantile_" + self.est.output_distribution)

            self.original_columns = names
        return self

    def transform(self, X, y=None, copy=None):
        self.index = X.index

        # peform apply learned transformation to input dataset
        X = self.est.transform(X)

        # if data is sparse, convert to dense
        if sparse.issparse(X):
            X = X.todense()

        # create new Pandas DataFrame with new column names
        X = pd.DataFrame(X, columns=self.original_columns, index=self.index)
        return X

class PandasFeatureUnion(FeatureUnion):
    """
    Documentation:

        ---
        Description:
            Modified version of sklearn's FeatureUnion class that outputs a Pandas DataFrame rather
            than a two dimensional array.
    """

    def fit_transform(self, X, y=None, **fit_params):

        # preserve mlm_dtypes if it exists
        try:
            self.meta_mlm_dtypes = X.mlm_dtypes
            self.no_meta_mlm_dtypes = False
        except AttributeError:
            self.no_meta_mlm_dtypes = True
            pass

        self._validate_transformers()
        result = Parallel(n_jobs=self.n_jobs)(
            delayed(_fit_transform_one)(
                transformer=trans,
                X=X,
                y=y,
                weight=weight,
                **fit_params)
            for name, trans, weight in self._iter())

        if not result:
            # All transformers are None
            return np.zeros((X.shape[0], 0))
        Xs, transformers = zip(*result)
        self._update_transformer_list(transformers)
        if any(sparse.issparse(f) for f in Xs):
            Xs = sparse.hstack(Xs).tocsr()
        else:
            Xs = self.merge_dataframes_by_column(Xs)

        if not self.no_meta_mlm_dtypes:
            Xs = Xs.loc[:, ~Xs.columns.duplicated()]
            Xs = PreserveMetaData(Xs)
            Xs.mlm_dtypes = self.meta_mlm_dtypes

            # reset dtype for any columns that were turned into object columns
            for mlm_dtype in Xs.mlm_dtypes.keys():
                for column in Xs.mlm_dtypes[mlm_dtype]:
                    try:
                        if is_object_dtype(Xs[column]):
                            if mlm_dtype == "boolean":
                                Xs[column] = Xs[column].astype("boolean")
                            elif mlm_dtype == "continuous":
                                Xs[column] = Xs[column].astype("float64")
                            elif mlm_dtype == "category":
                                Xs[column] = Xs[column].astype("category")
                            elif mlm_dtype == "count":
                                Xs[column] = Xs[column].astype("int64")
                            elif mlm_dtype == "date":
                                Xs[column] = Xs[column].astype("datetime64[ns]")
                            elif mlm_dtype == "nominal":
                                Xs[column] = Xs[column].astype("category")
                            elif mlm_dtype == "ordinal":
                                Xs[column] = Xs[column].astype("category")
                    except KeyError:
                        continue

        return Xs

    def merge_dataframes_by_column(self, Xs):
        return pd.concat(Xs, axis="columns", copy=False)

    def transform(self, X):
        Xs = Parallel(n_jobs=self.n_jobs)(
            delayed(_transform_one)(
                transformer=trans,
                X=X,
                y=None,
                weight=weight)
            for name, trans, weight in self._iter())
        if not Xs:
            # All transformers are None
            return np.zeros((X.shape[0], 0))
        if any(sparse.issparse(f) for f in Xs):
            Xs = sparse.hstack(Xs).tocsr()
        else:
            Xs = self.merge_dataframes_by_column(Xs)

        # metadata attributes
        if not self.no_meta_mlm_dtypes:
            Xs = Xs.loc[:, ~Xs.columns.duplicated()]
            Xs = PreserveMetaData(Xs)
            Xs.mlm_dtypes = self.meta_mlm_dtypes

            # reset dtype for any columns that were turned into object columns
            for mlm_dtype in Xs.mlm_dtypes.keys():
                for column in Xs.mlm_dtypes[mlm_dtype]:
                    try:
                        if is_object_dtype(Xs[column]):
                            if mlm_dtype == "boolean":
                                Xs[column] = Xs[column].astype("boolean")
                            elif mlm_dtype == "continuous":
                                Xs[column] = Xs[column].astype("float64")
                            elif mlm_dtype == "category":
                                Xs[column] = Xs[column].astype("category")
                            elif mlm_dtype == "count":
                                Xs[column] = Xs[column].astype("int64")
                            elif mlm_dtype == "date":
                                Xs[column] = Xs[column].astype("datetime64[ns]")
                            elif mlm_dtype == "nominal":
                                Xs[column] = Xs[column].astype("category")
                            elif mlm_dtype == "ordinal":
                                Xs[column] = Xs[column].astype("category")
                    except KeyError:
                        continue

        return Xs

class KFoldEncoder(BaseEstimator, TransformerMixin):
    """
    Documentation:

        ---
        Description:
            Perform KFold encoding using the provided cross-validation object and encoder. KFold
            encoding prevents leakage when using encoders that incorporate the target data. KFoldEncoder
            applies a procedure where the encoder fits to the training data and then applies the encoding
            on the held out dataset. This process repeats for each fold, per the setup of the cross-validation
            object, until all observations have been encoded in this manner.

        ---
        Parameters
            target : Pandas Series
                Pandas Series containing target values
            cv : sklearn cross-validation object
                Object to use when applying cross-validation to dataset
            encoder : encoder class
                Class to use when encoding features
            transform_train : boolean, default=False
                Controls whether KFoldEncoder's transform method runs in training mode or impute mode.
                Defaults to False so that when KFoldEncoder is first utilized, it runs in training mode
                so that it learns the encoding values. Once these values are learned, transform_train is
                automatically set to True so that subsequent calls of the transform method apply the
                previously learned encoding values.
    """
    def __init__(self, target, cv, encoder):
        self.target = target
        self.cv = cv
        self.encoder = encoder
        self.transform_train = False

    def fit(self, X, y=None):
        self.columns = X.columns
        self.stats = {}

        # capture name of encoder for column name
        encoder_name = self.encoder.__module__.split(".")
        if encoder_name[0] == "category_encoders":
            self.column_suffix = "_" + encoder_name[1]

        # map column suffix to desired name
        if self.column_suffix == "_woe":
            self.column_suffix = "_woe_encoded"
        if self.column_suffix == "_cat_boost":
            self.column_suffix = "_catboost_encoded"
        elif self.column_suffix == "_target_encoder":
            self.column_suffix = "_target_encoded"

        return self

    def transform(self, X, y=None):
        if not self.transform_train:
            # combine input columns and target
            X = X.merge(self.target, left_index=True, right_index=True)

            # add empty columns to input dataset and set as number
            for column in self.columns:
                X[column + self.column_suffix] = np.nan
                X[column + self.column_suffix] = pd.to_numeric(X[column + self.column_suffix])

            # iterate through cv indices
            for train_ix, valid_ix in self.cv.split(X):
                X_train, X_valid = X.iloc[train_ix], X.iloc[valid_ix]

                for column in self.columns:
                    enc = self.encoder(cols=[column])

                    # fit on train
                    X_train[column + self.column_suffix] = enc.fit_transform(X_train[column], X_train[self.target.name])
                    X_valid[column + self.column_suffix] = enc.transform(X_valid[column])
                    X.loc[X.index[valid_ix], column + self.column_suffix] = X_valid[column + self.column_suffix].map(X_valid[column + self.column_suffix]).fillna(X_valid[column + self.column_suffix])

            # fill nan's with column mean. rarely needed - only when column is severely imbalanced
            for column in X.columns:
                if X[column].isnull().sum() > 0:
                    X[column] = X[column].fillna(X[column].mean())

            # collect average values for transformation of unseen data
            for column in self.columns:
                self.stats[column] = X.groupby(column)[column + self.column_suffix].mean()

            # flip transform_train switch to indicate that fitting has occurred
            # ensure future transform calls will go to the else branch of the conditional
            self.transform_train = True

            # drop target column
            X = X.drop(self.target.name, axis=1)

        else:
            # iterate through all columns in the summary stats dict
            for column in self.stats.keys():

                # create empty colunn
                X[column + self.column_suffix] = np.nan
                X[column + self.column_suffix] = pd.to_numeric(X[column + self.column_suffix])

                # fill in empty column with mean values grouped by target
                X[column + self.column_suffix] = np.where(
                    X[column + self.column_suffix].isnull(),
                    X[column].map(self.stats[column]),
                    X[column + self.column_suffix],
                )
        return X

class DualTransformer(TransformerMixin, BaseEstimator):
    """
    Documentation:

        ---
        Description:
            Performs Yeo-Johnson transformation on all specified features. Also performs Box-Cox transformation.
            If the minimum value is greater than 0, standard Box-Cox transformation is performed. If the minimum
            value in a feature is 0, Box-Cox + 1 transformation is performed. If the minimum value in a feature is
            less than 0, Box-Cox + 1 + absolute value of the minimum value is performed. Each transformation
            process automatically determines the optimal lambda value. These values are stored for transformation
            of validation data.

            This method adds additional columns rather than updating existing columns in-place.

        ---
        Parameters:
            yeojohnson : bool, default=True
                Conditional controlling whether yeo-johnson transformation is applied to input data.
            boxcox : bool, default=True
                Conditional controlling whether Box-Cox transformation is applied to input data.

        --
        Returns:
            X : array
                Yeo-Johnson and Box-Coxtransformed input data.
    """

    def __init__(self, yeojohnson=True, boxcox=True):
        self.yeojohnson = yeojohnson
        self.boxcox = boxcox

    def fit(self, X, y=None):

        # capture original column names
        self.original_columns = X.columns

        ## Yeo-Johnson
        if self.yeojohnson:

            # empty dicitonary for capturing feature/lambda pairs
            self.yj_lambdas_dict_ = {}

            # collect lambas using sklearn PowerTransformer
            yj = PowerTransformer(method="yeo-johnson")
            yj_lambdas = yj.fit(X).lambdas_

            # cycle through columns, add transformed column name and associated lambdas
            for ix, col in enumerate(X.columns):
                self.yj_lambdas_dict_[col] = yj_lambdas[ix]

        ## Box-Cox
        if self.boxcox:

            # empty dicitonaries for capturing feature/lambda pairs
            self.bc_lambdas_dict_ = {}
            self.bc_zero_lambdas_dict_ = {}
            self.bc_neg_lambdas_dict_ = {}

            for col in X.columns:

                # if minimum feature value is 0, do Box-Cox + 1
                feature_minimum = np.min(X[col].values)
                if feature_minimum == 0:

                    # perform Box-Cox transformation and return lambda
                    _, lmbda = stats.boxcox(X[col].values + 1, lmbda=None)

                    # add transformed column name and associated lambda to dictionary
                    self.bc_zero_lambdas_dict_[col] = lmbda

                # if minimum feature value is greater than 0, do standard Box-Cox
                elif feature_minimum > 0:

                    # perform Box-Cox transformation and return lambda
                    _, lmbda = stats.boxcox(X[col].values, lmbda=None)

                    # add transformed column name and associated lambda to dictionary
                    self.bc_lambdas_dict_[col] = lmbda

                # if minimum feature value is less than 0, do Box-Cox + 1 + absolute value of minimum
                else:

                    # perform Box-Cox transformation and return lambda
                    _, lmbda = stats.boxcox(X[col].values + np.abs(feature_minimum) + 1, lmbda=None)

                    # add transformed column name and associated lambda to dictionary
                    self.bc_neg_lambdas_dict_[col] = lmbda
        return self

    def transform(self, X):
        # yeo-johnson
        if self.yeojohnson:

            # apply learned Yeo-Johnson transformation
            for col in self.yj_lambdas_dict_.keys():
                X[col + "_YeoJohnson"] = stats.yeojohnson(
                    X[col].values, lmbda=self.yj_lambdas_dict_[col]
                )

        # Box-Cox
        if self.boxcox:

            # apply learned Box-Cox transformation
            for col in self.bc_zero_lambdas_dict_.keys():
                try:
                    X[col + "_BoxCox"] = stats.boxcox(
                        X[col].values + 1, lmbda=self.bc_zero_lambdas_dict_[col]
                    )
                except ValueError:
                    X[col + "_BoxCox"] = 0.

            # apply learned Box-Cox transformation
            for col in self.bc_neg_lambdas_dict_.keys():
                try:
                    X[col + "_BoxCox"] = stats.boxcox(
                        X[col].values + np.abs(np.min(X[col].values)) + 1, lmbda=self.bc_neg_lambdas_dict_[col]
                    )
                except ValueError:
                    X[col + "_BoxCox"] = 0.

            # apply learned Box-Cox transformation
            for col in self.bc_lambdas_dict_.keys():
                try:
                    X[col + "_BoxCox"] = stats.boxcox(
                        X[col].values, lmbda=self.bc_lambdas_dict_[col]
                    )
                except ValueError:
                    X[col + "_BoxCox"] = 0.
        return X

def skew_summary(self, data=None, columns=None):
    """
    Documentation:

        ---
        Description:
            Displays Pandas DataFrame summarizing the skew of each numeric variable. Also calculates
            the percent of a column that is zero.

        ---
        Parameters:
            data : Pandas DataFrame, default=None
                Pandas DataFrame containing independent variables. If left as none,
                The feature dataset provided to Machine during instantiation is used.
            columns : list of strings, default=None
                List containing string names of columns. If left as none, the value associated
                with self.data.mlm_dtypes["continuous"] will be used as the column list.
    """
    # use data columns provided during instantiation if None
    if data is None:
        data = self.data
    # use mlm_dtypes["continuous"] columns provided during instantiation if None
    if columns is None:
        columns = self.data.mlm_dtypes["continuous"]

    # calculate skew for each column, dropping nulls and sorting on skew descending
    skewness = (
        data[columns]
        .apply(lambda X: stats.skew(X.dropna()))
        .sort_values(ascending=False)
    )

    # store results in DataFrame
    skewness = pd.DataFrame({"Skew": skewness})

    # add column describing percent of values that are zero
    skewness["Percent zero"] = np.nan
    for col in columns:

        try:
            skewness.loc[col]["Percent zero"] = data[data[col] == 0][
                col
            ].value_counts() / len(data)
        except ValueError:
            skewness.loc[col]["Percent zero"] = 0.0
    skewness = skewness.sort_values(["Skew"])

    return skewness

def missing_summary(self, data=None):
    """
    Documentation:

        ---
        Description:
            Displays Pandas DataFrame summarizing the missingness of each variable.

        ---
        Parameters:
            data : Pandas DataFrame, default=None
                Pandas DataFrame containing independent variables. If left as none,
                the feature dataset provided to Machine during instantiation is used.
    """
    # use data provided during instantiation if None
    if data is None:
        data = self.data

    # calculate missing data statistics
    total_missing = data.isnull().sum()
    percent_missing = data.isnull().sum() / len(data) * 100

    # create DataFrame using missingness data
    percent_missing = pd.DataFrame(
        {"Total missing": total_missing, "Percent missing": percent_missing}
    )

    # limit DataFrame to rows where "Percent missing" is greater than 0, sort descending
    # by "Percent missing"
    percent_missing = percent_missing[
        ~(percent_missing["Percent missing"].isna())
        & (percent_missing["Percent missing"] > 0)
    ].sort_values(["Percent missing"], ascending=False)

    return percent_missing

def unique_category_levels(self, data=None):
    """
    Documentation:

        ---
        Description:
            Print the unique values within each category column as define by mlm_dtypes["category"].

        ---
        Parameters:
            data : Pandas DataFrame, default=None
                Pandas DataFrame containing independent variables. If left as none,
                the feature dataset provided to Machine during instantiation is used.
    """

    # use data/ columns provided during instantiation if None
    if data is None:
        data = self.data

    # print unique values in each category columns
    for column in data.mlm_dtypes["category"]:
        print(column, "\t", np.unique(data[column]))

def compare_train_valid_levels(self, train_data, validation_data):
    """
    Documentation:

        ---
        Description:
            displays Pandas DataFrame summarizing the skew of each number variable. Also summarizes
            the percent of a column that has a value of zero.

        ---
        Parameters:
            train_data : Pandas DataFrame
                Pandas DataFrame containing training data.
            validation_data : Pandas DataFrame
                Pandas DataFrame containing validation data.
    """
    # creat counter to track how many features have differing amounts of unique values
    # between the input training and validation datasets
    counter = 0
    for col in train_data.mlm_dtypes["category"]:

        # capture unique feature values in both training and validation datasets
        train_values = train_data[col].unique()
        valid_values = validation_data[col].unique()

        # determine if categories exist in the training data and not the validation data
        train_diff = set(train_values) - set(valid_values)

        # determine if categories exist in the validation data and not the training data
        valid_diff = set(valid_values) - set(train_values)

        # if there is a discrepancy in the presence of category values in the training
        # and/or validation datasets, print summary information
        if len(train_diff) > 0 or len(valid_diff) > 0:
            print("\n\n*** " + col)
            print("Value present in training data, not in validation data")
            print(train_diff)
            print("Value present in validation data, not in training data")
            print(valid_diff)
            counter+=1

    # if all levels present in both datasets
    if counter == 0:
        print("All levels in all category columns present in both datasets.")

def missing_column_compare(self, validation_data, train_data=None):
    """
    Documentation:

        ---
        Description:
            Compares the columns that contain missing data in the training dataset
            to the columns that contain missing data in the validation dataset. Prints
            a summary of the disjunction.

        ---
        Parameters:
            validation_data : Pandas DataFrame
                Pandas DataFrame containing validation data.
            train_data : Pandas DataFrame, default=None
                Pandas DataFrame containing training data. If no value is passed, Method
                will default to using the Machine object's data attribute.
    """
    # use data provided during instantiation if None
    if train_data is None:
        train_data = self.data

    # return DataFrame describing number of values missing within each feature in the
    # training data
    train_missing = train_data.isnull().sum()

    # limit to feature that have greater than 0 missing values
    train_missing = train_missing[train_missing > 0].index

    # return DataFrame describing number of values missing within each feature in the
    # validation data
    validation_missing = validation_data.isnull().sum()

    # limit to feature that have greater than 0 missing values
    validation_missing = validation_missing[validation_missing > 0].index

    # print bi-direction summary of which columns are missing in the training dataset
    # versus the validation dataset
    print("Feature has missing values in validation data, not training data.")
    print(set(validation_missing) - set(train_missing))
    print("")
    print("Feature has missing values in training data, not validation data.")
    print(set(train_missing) - set(validation_missing))

class PreserveMetaData(pd.DataFrame):

    _metadata = ["mlm_dtypes"]

    @property
    def _constructor(self):
        return PreserveMetaData


import pandas as pd
from pandas.api.types import is_numeric_dtype, is_object_dtype
import numpy as np

import seaborn as sns
import squarify
import matplotlib.pyplot as plt
import matplotlib.ticker as tkr
import matplotlib.cm
from matplotlib.colors import ListedColormap, LinearSegmentedColormap

from statsmodels.stats.weightstats import ztest
from statsmodels.stats.proportion import proportions_ztest
from scipy import stats

from IPython.display import display_html

import os
import sys

from prettierplot.plotter import PrettierPlot
from prettierplot import style



def eda_cat_target_cat_feat(self, feature, level_count_cap=50, color_map="viridis", legend_labels=None,
                            chart_scale=15):
    """
    Documentation:

        ---
        Description:
            Creates exploratory data visualizations and statistical summaries for a category feature
            in the context of a categorical target.

        ---
        Parameters:
            feature : str
                Feature to visualize.
            level_count_cap : int, default=50
                Maximum number of unique levels in feature. If the number of levels exceeds the
                cap, then no visualization panel is produced.
            color_map : str specifying built-in matplotlib colormap, default="viridis"
                Color map applied to plots.
            legend_labels : list, default=None
                Class labels displayed in plot legend.
            chart_scale : int or float, default=15
                Controls size and proportions of chart and chart elements. Higher value creates
                larger plots and increases visual elements proportionally.
    """

    # if number of unique levels in feature is less than specified level_count_cap
    if (len(np.unique(self.data[self.data[feature].notnull()][feature].values)) < level_count_cap):

        ### data summaries
        ## feature summary
        # create empty DataFrame
        uni_summ_df = pd.DataFrame(columns=[feature, "Count", "Proportion"])

        # capture unique values and count of those unique values
        unique_vals, unique_counts = np.unique(
            self.data[self.data[feature].notnull()][feature], return_counts=True
        )

        # append each unique value, count and proportion to DataFrame
        for i, j in zip(unique_vals, unique_counts):
            uni_summ_df = uni_summ_df.append(
                {
                    feature: i,
                    "Count": j,
                    "Proportion": j / np.sum(unique_counts) * 100,
                },
                ignore_index=True,
            )

        # sort DataFrame by "Proportion", descending
        uni_summ_df = uni_summ_df.sort_values(by=["Proportion"], ascending=False)

        # set values to int dtype where applicable to optimize
        uni_summ_df["Count"] = uni_summ_df["Count"].astype("int64")
        if is_numeric_dtype(uni_summ_df[feature]):
            uni_summ_df[feature] = uni_summ_df[feature].astype("int64")

        ## feature vs. target summary
        # combine feature column and target
        bi_df = pd.concat([self.data[feature], self.target], axis=1)

        # remove any rows with nulls
        bi_df = bi_df[bi_df[feature].notnull()]

        # groupby category feature and count the occurrences of target classes
        # for each level in category
        bi_summ_df = (
            bi_df.groupby([feature] + [self.target.name])
            .size()
            .reset_index()
            .pivot(columns=self.target.name, index=feature, values=0)
        )

        # overwrite DataFrame index with actual class labels if provided
        bi_summ_df.columns = pd.Index(legend_labels) if legend_labels is not None else pd.Index([i for i in bi_summ_df.columns.tolist()])
        bi_summ_df.reset_index(inplace=True)

        # fill nan's with zero
        fill_columns = bi_summ_df.iloc[:,2:].columns
        bi_summ_df[fill_columns] = bi_summ_df[fill_columns].fillna(0)

        # set values to int dtype where applicable to optimize displayed DataFrame
        for column in bi_summ_df.columns:
            try:
                bi_summ_df[column] = bi_summ_df[column].astype(np.int)
            except ValueError:
                bi_summ_df[column] = bi_summ_df[column]

        ## proportion by category summary
        # combine feature column and target
        prop_df = pd.concat([self.data[feature], self.target], axis=1)

        # remove any rows with nulls
        prop_df = prop_df[prop_df[feature].notnull()]

        # calculate percent of 100 by class label
        prop_df = prop_df.groupby([feature, self.target.name]).agg({self.target.name : {"count"}})
        prop_df = prop_df.groupby(level=0).apply(lambda x: 100 * x / float(x.sum()))
        prop_df = prop_df.reset_index()

        multiIndex = prop_df.columns
        singleIndex = [i[0] for i in multiIndex.tolist()]
        singleIndex[-1] = "Count"
        prop_df.columns = singleIndex
        prop_df = prop_df.reset_index(drop=True)

        prop_df = pd.pivot_table(prop_df, values=["Count"], columns=[feature], index=[self.target.name], aggfunc={"Count": np.mean})
        prop_df = prop_df.reset_index(drop=True)

        multiIndex = prop_df.columns
        singleIndex = []

        for column in multiIndex.tolist():
            try:
                singleIndex.append(int(column[1]))
            except ValueError:
                singleIndex.append(column[1])

        prop_df.columns = singleIndex
        prop_df = prop_df.reset_index(drop=True)

        # insert column to DataFrame with actual class labels if provided, otherwise use raw class labels in target
        prop_df.insert(loc=0, column="Class", value=legend_labels if legend_labels is not None else np.unique(self.target))

        # fill nan's with zero
        fill_columns = prop_df.iloc[:,:].columns
        prop_df[fill_columns] = prop_df[fill_columns].fillna(0)

        # if there are only two class labels, perform z-test/t-test
        if len(np.unique(bi_df[bi_df[feature].notnull()][feature])) == 2:

            # total observations
            total_obs1 = bi_df[(bi_df[feature] == np.unique(bi_df[feature])[0])][
                feature
            ].shape[0]
            total_obs2 = bi_df[(bi_df[feature] == np.unique(bi_df[feature])[1])][
                feature
            ].shape[0]

            # total positive observations
            pos_obs1 = bi_df[
                (bi_df[feature] == np.unique(bi_df[feature])[0])
                & (bi_df[self.target.name] == 1)
            ][feature].shape[0]
            pos_obs2 = bi_df[
                (bi_df[feature] == np.unique(bi_df[feature])[1])
                & (bi_df[self.target.name] == 1)
            ][feature].shape[0]

            # perform z-test, return z-statistic and p-value
            z, p_val = proportions_ztest(
                count=(pos_obs1, pos_obs2), nobs=(total_obs1, total_obs2)
            )

            # add z-statistic and p-value to DataFrame
            stat_test_df = pd.DataFrame(
                data=[{"z-test statistic": z, "p-value": p_val}],
                columns=["z-test statistic", "p-value"],
                index=[feature],
            ).round(4)

            # display summary tables
            self.df_side_by_side(
                dfs=(uni_summ_df, bi_summ_df, prop_df, stat_test_df),
                names=["Feature summary", "Feature vs. target summary", "Target proportion", "Statistical test",],
            )
            if "percent_positive" in bi_summ_df:
                bi_summ_df = bi_summ_df.drop(["percent_positive"], axis=1)

        else:
            # display summary tables
            self.df_side_by_side(
                dfs=(uni_summ_df, bi_summ_df, prop_df),
                names=["Feature summary", "Feature vs. target summary", "Target proportion"],
            )
            if "percent_positive" in bi_summ_df:
                bi_summ_df = bi_summ_df.drop(["percent_positive"], axis=1)

        ### visualizations
        # set label rotation angle
        len_unique_val = len(unique_vals)
        avg_len_unique_val = sum(map(len, str(unique_vals))) / len(unique_vals)
        if len_unique_val <= 4 and avg_len_unique_val <= 12:
            rotation = 0
        elif len_unique_val >= 5 and len_unique_val <= 8 and avg_len_unique_val <= 8:
            rotation = 0
        elif len_unique_val >= 9 and len_unique_val <= 14 and avg_len_unique_val <= 4:
            rotation = 0
        else:
            rotation = 90

        # create prettierplot object
        p = PrettierPlot(chart_scale=chart_scale, plot_orientation="wide_narrow")

        # add canvas to prettierplot object
        ax = p.make_canvas(title="Category counts\n* {}".format(feature), position=131, title_scale=0.82)

        # add treemap to canvas
        p.tree_map(
            counts=uni_summ_df["Count"].values,
            labels=uni_summ_df[feature].values,
            colors=style.color_gen(name=color_map, num=len(uni_summ_df[feature].values)),
            alpha=0.8,
            ax=ax,
        )

        # add canvas to prettierplot object
        ax = p.make_canvas(title="Category counts by target\n* {}".format(feature), position=132)

        # add faceted categorical plot to canvas
        p.facet_cat(
            df=bi_summ_df,
            feature=feature,
            label_rotate=rotation,
            color_map=color_map,
            bbox=(1.0, 1.15),
            alpha=0.8,
            legend_labels=legend_labels,
            x_units=None,
            ax=ax,
        )

        # add canvas to prettierplot object
        ax = p.make_canvas(title="Target proportion by category\n* {}".format(feature), position=133)

        # add stacked bar chart to canvas
        p.stacked_bar_h(
            df=prop_df.drop("Class", axis=1),
            bbox=(1.0, 1.15),
            legend_labels=legend_labels,
            color_map=color_map,
            alpha=0.8,
            ax=ax,
        )

        plt.show()


def eda_cat_target_num_feat(self, feature, color_map="viridis", outliers_out_of_scope=None, legend_labels=None,
                            chart_scale=15):
    """
    Documentation:

        ---
        Description:
            Creates exploratory data visualizations and statistical summaries for a number
            feature in the context of a categorical target.

        ---
        Parameters:
            feature : str
                Feature to visualize.
            color_map : str specifying built-in matplotlib colormap, default="viridis"
                Color map applied to plots.
            outliers_out_of_scope : boolean, float or int, default=None
                Truncates the x-axis upper limit so that outliers are out of scope of the visualization.
                The x-axis upper limit is reset to the maximum non-outlier value.

                To identify outliers, the IQR is calculated, and values that are below the first quartile
                minus the IQR, or above the third quarterile plus the IQR are designated as outliers. If True
                is passed as a value, the IQR that is subtracted/added is multiplied by 5. If a float or int is
                passed, the IQR is multiplied by that value. Higher values increase how extremem values need
                to be to be identified as outliers.
            legend_labels : list, default=None
                Class labels displayed in plot legend.
            chart_scale : int or float, default=15
                Controls size and proportions of chart and chart elements. Higher value creates larger plots
                and increases visual elements proportionally.
    """
    ### data summaries
    ## bivariate roll_up table
    # combine feature column and target
    bi_df = pd.concat([self.data[feature], self.target], axis=1)

    # remove any rows with nulls
    bi_df = bi_df[bi_df[feature].notnull()]

    # bivariate summary statistics
    bi_summ_stats_df = pd.DataFrame(
        columns=["Class", "Count", "Proportion", "Mean", "StdDev"]
    )

    # for each unique class label
    for labl in np.unique(self.target):

        # get feature values associated with single class label
        feature_slice = bi_df[bi_df[self.target.name] == labl][feature]

        # append summary statistics for feature values associated with class label
        bi_summ_stats_df = bi_summ_stats_df.append(
            {
                "Class": labl,
                "Count": len(feature_slice),
                "Proportion": len(feature_slice) / len(bi_df[feature]) * 100,
                "Mean": np.mean(feature_slice),
                "StdDev": np.std(feature_slice),
            },
            ignore_index=True,
        )

    # apply custom legend labels, or set dtype to int if column values are numeric
    if legend_labels is not None:
        bi_summ_stats_df["Class"] = legend_labels
    elif is_numeric_dtype(bi_summ_stats_df["Class"]):
        bi_summ_stats_df["Class"] = bi_summ_stats_df["Class"].astype(np.int)


    ## Feature summary
    describe_df = pd.DataFrame(bi_df[feature].describe()).reset_index()

    # add missing percentage
    describe_df = describe_df.append(
        {
            "index": "missing",
            feature: np.round(self.data.shape[0] - bi_df[feature].shape[0], 5),
        },
        ignore_index=True,
    )

    # add skew
    describe_df = describe_df.append(
        {
            "index": "skew",
            feature: np.round(stats.skew(bi_df[feature].values, nan_policy="omit"), 5),
        },
        ignore_index=True,
    )
    # add kurtosis
    describe_df = describe_df.append(
        {
            "index": "kurtosis",
            feature: stats.kurtosis(bi_df[feature].values, nan_policy="omit"),
        },
        ignore_index=True,
    )
    describe_df = describe_df.rename(columns={"index": ""})

    # execute z-test or t-test
    if len(np.unique(self.target)) == 2:
        s1 = bi_df[
            (bi_df[self.target.name] == bi_df[self.target.name].unique()[0])
        ][feature]
        s2 = bi_df[
            (bi_df[self.target.name] == bi_df[self.target.name].unique()[1])
        ][feature]
        if len(s1) > 30 and len(s2) > 30:

            # perform z-test, return z-statistic and p-value
            z, p_val = ztest(s1, s2)

            # add z-statistic and p-value to DataFrame
            stat_test_df = pd.DataFrame(
                data=[{"z-test statistic": z, "p-value": p_val}],
                columns=["z-test statistic", "p-value"],
                index=[feature],
            ).round(4)
        else:
            # perform t-test, return t-score and p-value
            t, p_val = stats.ttest_ind(s1, s2)

            # add t-statistic and p-value to DataFrame
            stat_test_df = pd.DataFrame(
                data=[{"t-test statistic": t, "p-value": p_val}],
                columns=["t-test statistic", "p-value"],
                index=[feature],
            ).round(4)

        # display summary tables
        self.df_side_by_side(
            dfs=(describe_df, bi_summ_stats_df, stat_test_df),
            names=["Feature summary", "Feature vs. target summary", "Statistical test"],
        )
    else:

        # display summary tables
        self.df_side_by_side(
            dfs=(describe_df, bi_summ_stats_df),
            names=["Feature summary", "Feature vs. target summary"],
        )

    ### visualizations
    # create prettierplot object
    p = PrettierPlot(chart_scale=chart_scale, plot_orientation="wide_standard")

    # if boolean is passed to outliers_out_of_scope
    if isinstance(outliers_out_of_scope, bool):
        # if outliers_out_of_scope = True
        if outliers_out_of_scope:

            # identify outliers using IQR method and an IQR step of 5
            outliers = self.outlier_IQR(self.data[feature], iqr_step=5)

            # reset x-axis minimum and maximum
            x_axis_min = self.data[feature].drop(index=outliers).min()
            x_axis_max = self.data[feature].drop(index=outliers).max()
    # if outliers_out_of_scope is a float or int
    elif isinstance(outliers_out_of_scope, float) or isinstance(outliers_out_of_scope, int):
        # identify outliers using IQR method and an IQR step equal to the float/int passed
        outliers = self.outlier_IQR(self.data[feature], iqr_step=outliers_out_of_scope)

        # reset x-axis minimum and maximum
        x_axis_min = self.data[feature].drop(index=outliers).min()
        x_axis_max = self.data[feature].drop(index=outliers).max()

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="Feature distribution\n* {}".format(feature),
        title_scale=0.85,
        position=221,
    )

    ## dynamically determine precision of x-units
    # capture min and max feature values
    dist_min = bi_df[feature].values.min()
    dist_max = bi_df[feature].values.max()

    # determine x-units precision based on min and max values in feature
    if -3 < dist_min < 3 and -3 < dist_max < 3 and dist_max/dist_min < 10:
        x_units = "fff"
    elif -30 < dist_min < 30 and -30 < dist_max < 30 and dist_max/dist_min < 3:
        x_units = "fff"
    elif -5 < dist_min < 5 and -5 < dist_max < 5 and dist_max/dist_min < 10:
        x_units = "ff"
    elif -90 < dist_min < 90 and -90 < dist_max < 90 and dist_max/dist_min < 5:
        x_units = "ff"
    else:
        x_units = "f"

    # add distribution plot to canvas
    p.dist_plot(
        bi_df[feature].values,
        color=style.style_grey,
        y_units="f",
        x_units=x_units,
        ax=ax,
    )

    # optionally reset x-axis limits
    if outliers_out_of_scope is not None:
        plt.xlim(x_axis_min, x_axis_max)

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="Probability plot\n* {}".format(feature),
        title_scale=0.85,
        position=222,
    )

    # add QQ / probability plot to canvas
    p.prob_plot(
        x=bi_df[feature].values,
        plot=ax,
    )

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="Distribution by class\n* {}".format(feature),
        title_scale=0.85,
        position=223,
    )

    ## dynamically determine precision of x-units
    # capture min and max feature values
    dist_min = bi_df[feature].values.min()
    dist_max = bi_df[feature].values.max()

    # determine x-units precision based on min and max values in feature
    if -3 < dist_min < 3 and -3 < dist_max < 3 and dist_max/dist_min < 10:
        x_units = "fff"
    elif -30 < dist_min < 30 and -30 < dist_max < 30 and dist_max/dist_min < 3:
        x_units = "fff"
    elif -5 < dist_min < 5 and -5 < dist_max < 5 and dist_max/dist_min < 10:
        x_units = "ff"
    elif -90 < dist_min < 90 and -90 < dist_max < 90 and dist_max/dist_min < 5:
        x_units = "ff"
    else:
        x_units = "f"

    # generate color list
    color_list = style.color_gen(name=color_map, num=len(np.unique(self.target)))

    # add one distribution plot to canvas for each category class
    for ix, labl in enumerate(np.unique(bi_df[self.target.name].values)):
        p.dist_plot(
            bi_df[bi_df[self.target.name] == labl][feature].values,
            color=color_list[ix],
            y_units="f",
            x_units=x_units,
            legend_labels=legend_labels if legend_labels is not None else np.arange(len(np.unique(self.target))),
            alpha=0.4,
            bbox=(1.0, 1.0),
            ax=ax,
        )

    # optionally reset x-axis limits
    if outliers_out_of_scope is not None:
        plt.xlim(x_axis_min, x_axis_max)

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="Boxplot by class\n* {}".format(feature),
        title_scale=0.85,
        position=224,
    )

    ## dynamically determine precision of x-units
    # capture min and max feature values
    dist_min = bi_df[feature].values.min()
    dist_max = bi_df[feature].values.max()

    # determine x-units precision based on min and max values in feature
    if -3 < dist_min < 3 and -3 < dist_max < 3 and dist_max/dist_min < 10:
        x_units = "fff"
    elif -30 < dist_min < 30 and -30 < dist_max < 30 and dist_max/dist_min < 3:
        x_units = "fff"
    elif -5 < dist_min < 5 and -5 < dist_max < 5 and dist_max/dist_min < 10:
        x_units = "ff"
    elif -90 < dist_min < 90 and -90 < dist_max < 90 and dist_max/dist_min < 5:
        x_units = "ff"
    else:
        x_units = "f"

    # add horizontal box plot to canvas
    p.box_plot_h(
        x=feature,
        y=self.target.name,
        data=bi_df,
        alpha=0.7,
        x_units=x_units,
        legend_labels=legend_labels,
        bbox=(1.2, 1.0),
        suppress_outliers=True,
        ax=ax
        )

    # optionally reset x-axis limits
    if outliers_out_of_scope is not None:
        plt.xlim(x_axis_min-(x_axis_min * 0.1), x_axis_max)

    # apply position adjustment to subplots
    plt.subplots_adjust(bottom=-0.1)

    plt.show()


def eda_num_target_num_feat(self, feature, color_map="viridis", chart_scale=15):
    """
    Documentation:

        ---
        Description:
            Produces exploratory data visualizations and statistical summaries for a numeric
            feature in the context of a numeric target.

        ---
        Parameters:
            feature : str
                Feature to visualize.
            color_map : str specifying built-in matplotlib colormap, default="viridis"
                Color map applied to plots.
            chart_scale : int or float, default=15
                Controls size and proportions of chart and chart elements. Higher value creates
                larger plots and increases visual elements proportionally.
    """
    ### data summaries
    ## feature summary
    # combine feature column and target
    bi_df = pd.concat([self.data[feature], self.target], axis=1)

    # remove any rows with nulls
    bi_df = bi_df[bi_df[feature].notnull()]

    # cast target as float
    bi_df[self.target.name] = bi_df[self.target.name].astype(float)

    # create summary statistic table
    describe_df = pd.DataFrame(bi_df[feature].describe()).reset_index()

    # add skew and kurtosis to describe_df
    describe_df = describe_df.append(
        {
            "index": "skew",
            feature: stats.skew(bi_df[feature].values, nan_policy="omit"),
        },
        ignore_index=True,
    )
    describe_df = describe_df.append(
        {
            "index": "kurtosis",
            feature: stats.kurtosis(bi_df[feature].values, nan_policy="omit"),
        },
        ignore_index=True,
    )
    describe_df = describe_df.rename(columns={"index": ""})

    # display summary tables
    display(describe_df)

    ### visualizations
    # create prettierplot object
    p = PrettierPlot(chart_scale=chart_scale, plot_orientation="wide_narrow")

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="Feature distribution\n* {}".format(feature), position=131, title_scale=1.2
    )

    # determine x-units precision based on magnitude of max value
    if -1 <= np.nanmax(bi_df[feature].values) <= 1:
        x_units = "fff"
    elif -10 <= np.nanmax(bi_df[feature].values) <= 10:
        x_units = "ff"
    else:
        x_units = "f"

    # determine y-units precision based on magnitude of max value
    if -1 <= np.nanmax(bi_df[feature].values) <= 1:
        y_units = "fff"
    elif -10 <= np.nanmax(bi_df[feature].values) <= 10:
        y_units = "ff"
    else:
        y_units = "f"

    # x rotation
    if -10000 < np.nanmax(bi_df[feature].values) < 10000:
        x_rotate = 0
    else:
        x_rotate = 45

    # add distribution plot to canvas
    p.dist_plot(
        bi_df[feature].values,
        color=style.style_grey,
        y_units=y_units,
        x_rotate=x_rotate,
        ax=ax,
    )

    # add canvas to prettierplot object
    ax = p.make_canvas(title="Probability plot\n* {}".format(feature), position=132)

    # add QQ / probability plot to canvas
    p.prob_plot(x=bi_df[feature].values, plot=ax)

    # add canvas to prettierplot object
    ax = p.make_canvas(
        title="Regression plot - feature vs. target\n* {}".format(feature),
        position=133,
        title_scale=1.5
        )

    # add regression plot to canvas
    p.reg_plot(
        x=feature,
        y=self.target.name,
        data=bi_df,
        x_jitter=0.1,
        x_rotate=x_rotate,
        x_units=x_units,
        y_units=y_units,
        ax=ax,
    )
    plt.show()


def eda_num_target_cat_feat(self, feature, level_count_cap=50, color_map="viridis", chart_scale=15):
    """
    Documentation:

        ---
        Description:
            Produces exploratory data visualizations and statistical summaries for a category
            feature in the context of a numeric target.

        ---
        Parameters:
            feature : str
                Feature to visualize.
            level_count_cap : int, default=50
                Maximum number of unique levels in feature. If the number of levels exceeds the
                cap then the feature is skipped.
            color_map : str specifying built-in matplotlib colormap, default="viridis"
                Color map applied to plots.
            chart_scale : int or float, default=15
                Controls size and proportions of chart and chart elements. Higher value creates
                larger plots and increases visual elements proportionally.
    """
    # if number of unique levels in feature is less than specified level_count_cap
    if (len(np.unique(self.data[self.data[feature].notnull()][feature].values)) < level_count_cap):

        ### data summaries
        ## feature summary
        # create empty DataFrame
        uni_summ_df = pd.DataFrame(columns=[feature, "Count", "Proportion"])

        # capture unique values and count of those unique values
        unique_vals, unique_counts = np.unique(
            self.data[self.data[feature].notnull()][feature], return_counts=True
        )

        # append each unique value, count and proportion to DataFrame
        for i, j in zip(unique_vals, unique_counts):
            uni_summ_df = uni_summ_df.append(
                {feature: i, "Count": j, "Proportion": j / np.sum(unique_counts) * 100},
                ignore_index=True,
            )

        # sort DataFrame by "Proportion", descending
        uni_summ_df = uni_summ_df.sort_values(by=["Proportion"], ascending=False)

        # set values to int dtype where applicable to optimize
        if is_numeric_dtype(uni_summ_df[feature]):
            uni_summ_df[feature] = uni_summ_df[feature].astype("int64")
        uni_summ_df["Count"] = uni_summ_df["Count"].astype("int64")

        ## feature vs. target summary
        # combine feature column and target
        bi_df = pd.concat([self.data[feature], self.target], axis=1)

        # remove any rows with nulls
        bi_df = bi_df[bi_df[feature].notnull()]

        # cast target as float
        bi_df[self.target.name] = bi_df[self.target.name].astype(float)

        # create pivot table of target summary statistics, grouping by category feature
        bi_summ_piv_df = pd.pivot_table(
            bi_df, index=feature, aggfunc={self.target.name: [np.nanmin, np.nanmax, np.nanmean, np.nanmedian, np.nanstd]}
        )
        multi_index = bi_summ_piv_df.columns
        single_index = pd.Index([i[1] for i in multi_index.tolist()])
        bi_summ_piv_df.columns = single_index
        bi_summ_piv_df.reset_index(inplace=True)
        bi_summ_piv_df = bi_summ_piv_df.rename(columns={
                                                "nanmin":"Min",
                                                "nanmax":"Max",
                                                "nanmean":"Mean",
                                                "nanmedian":"Median",
                                                "nanstd":"StdDev",
                                            }
                                        )
        # fill nan's with zero
        fill_columns = bi_summ_piv_df.iloc[:,1:].columns
        bi_summ_piv_df[fill_columns] = bi_summ_piv_df[fill_columns].fillna(0)

        # reorder column
        bi_summ_piv_df = bi_summ_piv_df[[feature,"Mean","Median","StdDev","Min","Max"]]

        # convert to int
        if is_numeric_dtype(bi_summ_piv_df[feature]):
            bi_summ_piv_df[feature] = bi_summ_piv_df[feature].astype("int64")

        # display summary tables
        self.df_side_by_side(
            dfs=(uni_summ_df, bi_summ_piv_df),
            names=["Feature summary", "Feature vs. target summary"],
        )

        ### visualizations
        # create prettierplot object
        p = PrettierPlot(chart_scale=chart_scale, plot_orientation="wide_narrow")

        # add canvas to prettierplot object
        ax = p.make_canvas(title="Category counts\n* {}".format(feature), position=131, title_scale=1.0)

        # add treemap to canvas
        p.tree_map(
            counts=uni_summ_df["Count"].values,
            labels=uni_summ_df[feature].values,
            colors=style.color_gen(name=color_map, num=len(uni_summ_df[feature].values)),
            alpha=0.8,
            ax=ax,
        )

        # add canvas to prettierplot object
        ax = p.make_canvas(title="Feature distribution\n* {}".format(feature), position=132)

        # error catching block for resorting labels
        try:
            sorted(unique_vals, key=int)
        except ValueError:
            pass
        else:
            # sort unique_vals/unique_counts for bar chart
            new_ix = [
                sorted(list(unique_vals), key=int).index(i) for i in list(unique_vals)
            ]
            unique_vals = np.array(sorted(list(unique_vals), key=int))
            unique_counts = np.array([y for x, y in sorted(zip(new_ix, unique_counts))])

            # sort temporary data frame for box plot
            bi_df[feature] = bi_df[feature].astype(int)

        # dynamically set rotation angle based on number unique values and maximum length of
        # category labels.
        len_unique_val = len(unique_vals)
        avg_len_unique_val = sum(map(len, str(unique_vals))) / len(unique_vals)
        if len_unique_val <= 4 and avg_len_unique_val <= 12:
            rotation = 0
        elif len_unique_val >= 5 and len_unique_val <= 8 and avg_len_unique_val <= 7.0:
            rotation = 0
        elif len_unique_val >= 9 and len_unique_val <= 14 and avg_len_unique_val <= 6:
            rotation = 0
        else:
            rotation = 30

        # represent x-axis tick labels as integers rather than floats
        x_values = list(map(str, unique_vals.tolist()))
        try:
            x_values = [int(float(x)) for x in x_values]
        except ValueError:
            pass

        # add bar chart to canvas
        p.bar_v(
            x=x_values,
            counts=unique_counts,
            label_rotate=rotation,
            color=style.style_grey,
            y_units="f",
            x_tick_wrap=True,
            ax=ax,
        )

        # hide every other label if total number of levels is greater than 40
        if len_unique_val > 40:
            n = 2
            [
                l.set_visible(False)
                for (i, l) in enumerate(ax.xaxis.get_ticklabels())
                if i % n != 0
            ]

        # add canvas to prettierplot object
        ax = p.make_canvas(
            title="Boxplot by category\n* {}".format(feature), position=133
        )

        ## dynamically determine precision of y-units
        # capture min and max feature values
        dist_min = bi_df[self.target.name].values.min()
        dist_max = bi_df[self.target.name].values.max()

        # determine y-units precision based on min and max values in feature
        if -3 < dist_min < 3 and -3 < dist_max < 3 and dist_max/dist_min < 10:
            y_units = "fff"
        elif -30 < dist_min < 30 and -30 < dist_max < 30 and dist_max/dist_min < 3:
            y_units = "fff"
        elif -5 < dist_min < 5 and -5 < dist_max < 5 and dist_max/dist_min < 10:
            y_units = "ff"
        elif -90 < dist_min < 90 and -90 < dist_max < 90 and dist_max/dist_min < 5:
            y_units = "ff"
        else:
            y_units = "f"

        # add vertical box plot to canvas
        p.box_plot_v(
            x=feature,
            y=self.target.name,
            data=bi_df.sort_values([feature]),
            color=matplotlib.cm.get_cmap(name=color_map),
            label_rotate=rotation,
            y_units=y_units,
            ax=ax,
        )

        # hide every other label if total number of levels is greater than 40
        if len_unique_val > 40:
            n = 2
            [
                l.set_visible(False)
                for (i, l) in enumerate(ax.xaxis.get_ticklabels())
                if i % n != 0
            ]

        plt.show()


def df_side_by_side(self, dfs, names=[]):
    """
    Documentation:

        ---
        Description:
            Helper function for displaying Pandas DataFrames side by side in a
            Jupyter Notebook.

        ---
        Parameters:
            dfs : list
                List of Pandas DataFrames to display.
            names : list, default=[]
                List of names to be displayed above Pandas DataFrames.
    """
    html_str = ""
    if names:
        html_str += (
            "<tr>"
            + "".join(f'<td style="text_align:center">{name}</td>' for name in names)
            + "</tr>"
        )
    html_str += (
        "<tr>"
        + "".join(
            f'<td style="vertical_align:top"> {df.to_html(index=False)}</td>'
            for df in dfs
        )
        + "</tr>"
    )
    html_str = f"<table>{html_str}</table>"
    html_str = html_str.replace("table", 'table style="display:inline"')
    display_html(html_str, raw=True)